'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.visualization.petri_net.common import visualize
from enum import Enum
from pm4py.util.constants import (
    PARAMETER_CONSTANT_ACTIVITY_KEY,
    PARAMETER_CONSTANT_TIMESTAMP_KEY,
)
from pm4py.objects.petri_net.obj import PetriNet, Marking
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
import graphviz


class Parameters(Enum):
    FORMAT = "format"
    DEBUG = "debug"
    RANKDIR = "set_rankdir"
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = PARAMETER_CONSTANT_TIMESTAMP_KEY
    AGGREGATION_MEASURE = "aggregationMeasure"
    FONT_SIZE = "font_size"


def apply(
    net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
    log: EventLog = None,
    aggregated_statistics=None,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> graphviz.Digraph:
    """
    Apply method for Petri net visualization (it calls the
    graphviz_visualization method)

    Parameters
    -----------
    net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    log
        (Optional) log
    aggregated_statistics
        Dictionary containing the frequency statistics
    parameters
        Algorithm parameters

    Returns
    -----------
    viz
        Graph object
    """
    # remove unused variables
    del log
    del aggregated_statistics
    return visualize.apply(
        net, initial_marking, final_marking, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from abc import ABC
from typing import Any, Optional, Dict, Generic, Tuple, List

from pm4py.algo.discovery.inductive.cuts.concurrency import (
    ConcurrencyCut,
    ConcurrencyCutUVCL,
    T,
)
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.objects.powl.obj import StrictPartialOrder


class POWLConcurrencyCut(ConcurrencyCut, ABC, Generic[T]):

    @classmethod
    def operator(
        cls, parameters: Optional[Dict[str, Any]] = None
    ) -> StrictPartialOrder:
        raise Exception("This function should not be called!")

    @classmethod
    def apply(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[StrictPartialOrder, List[T]]]:
        g = cls.holds(obj, parameters)
        if g is None:
            return g
        else:
            children = cls.project(obj, g, parameters)
            return StrictPartialOrder(children), children


class POWLConcurrencyCutUVCL(
    ConcurrencyCutUVCL, POWLConcurrencyCut[IMDataStructureUVCL]
):
    pass



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.process_tree import obj as pt_operator
from pm4py.util.regex import SharedObj, get_new_char

from pm4py.util import constants
import warnings

if constants.SHOW_INTERNAL_WARNINGS:
    warnings.warn("The regex package will be removed in a future release.")


def pt_to_regex(tree, rec_depth=0, shared_obj=None, parameters=None):
    """
    Transforms a process tree to a regular expression

    NB: The conversion is not yet working with trees containing an AND and/or an OR operator!

    Parameters
    ------------
    tree
        Process tree
    parameters
        Possible parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    if shared_obj is None:
        shared_obj = SharedObj()

    stru = ""

    if tree.operator is not None:
        contains_tau = (
            len(
                list(
                    child
                    for child in tree.children
                    if child.operator is None and child.label is None
                )
            )
            > 0
        )
        children_rep = []
        for child in tree.children:
            rep, shared_obj = pt_to_regex(
                child,
                rec_depth=rec_depth + 1,
                shared_obj=shared_obj,
                parameters=parameters,
            )
            children_rep.append(rep)
        if tree.operator == pt_operator.Operator.SEQUENCE:
            children_rep = [x for x in children_rep if x is not None]
            stru = "(" + "".join(children_rep) + ")"
        elif tree.operator == pt_operator.Operator.XOR:
            children_rep = [x for x in children_rep if x is not None]
            stru = "(" + "|".join(children_rep) + ")"
            if contains_tau:
                stru = "(" + stru + "?)"
        elif tree.operator == pt_operator.Operator.LOOP:
            children_rep = [x for x in children_rep if x is not None]
            if len(children_rep) == 1:
                stru = "(" + children_rep[0] + ")+"
            else:
                stru = "(" + "".join(children_rep) + ")*" + children_rep[0]
        elif tree.operator == pt_operator.Operator.PARALLEL:
            raise Exception(
                "the conversion is not yet working with trees containing an AND and/or an OR operator!"
            )
        elif tree.operator == pt_operator.Operator.OR:
            raise Exception(
                "the conversion is not yet working with trees containing an AND and/or an OR operator!"
            )

    elif tree.label is not None:
        if tree.label not in shared_obj.mapping_dictio:
            get_new_char(tree.label, shared_obj)
        stru = shared_obj.mapping_dictio[tree.label]
    elif tree.label is None:
        return None, shared_obj

    if rec_depth == 0:
        ret = "^" + stru + "$", shared_obj.mapping_dictio
        # print(ret)
        return ret

    return stru, shared_obj



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import uuid

from typing import Optional, Dict, Any, Union
from pm4py.objects.transition_system.obj import TransitionSystem
import graphviz

import tempfile

from graphviz import Digraph
from pm4py.util import exec_utils, constants
from enum import Enum


class Parameters(Enum):
    FORMAT = "format"
    BGCOLOR = "bgcolor"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def get_perc(total_events, arc_events):
    if total_events > 0:
        return (
            " "
            + str(total_events)
            + " / %.2f %%" % (100.0 * arc_events / total_events)
        )
    return " 0 / 0.00 %"


def apply(
    tsys: TransitionSystem,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> graphviz.Digraph:
    if parameters is None:
        parameters = {}

    image_format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, "png"
    )
    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR
    )

    enable_graph_title = exec_utils.get_param_value(
        Parameters.ENABLE_GRAPH_TITLE,
        parameters,
        constants.DEFAULT_ENABLE_GRAPH_TITLES,
    )
    graph_title = exec_utils.get_param_value(
        Parameters.GRAPH_TITLE, parameters, "Transition System"
    )

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    viz = Digraph(
        tsys.name,
        filename=filename.name,
        engine="dot",
        graph_attr={"bgcolor": bgcolor},
    )

    if enable_graph_title:
        viz.attr(
            label='<<FONT POINT-SIZE="20">' + graph_title + "</FONT>>",
            labelloc="top",
        )

    states_dictio = {}

    for s in tsys.states:
        node_uuid = str(uuid.uuid4())
        states_dictio[id(s)] = node_uuid

        sum_ingoing = 0
        sum_outgoing = 0

        for t in s.incoming:
            sum_ingoing += len(t.data["events"])

        for t in s.outgoing:
            sum_outgoing += len(t.data["events"])

        fillcolor = "white"

        if sum_ingoing != len(s.data["ingoing_events"]) or sum_outgoing != len(
            s.data["outgoing_events"]
        ):
            fillcolor = "red"

        taillabel = get_perc(sum_ingoing, len(s.data["ingoing_events"]))
        headlabel = get_perc(sum_outgoing, len(s.data["outgoing_events"]))

        label = "IN=" + taillabel + "\n" + str(s.name) + "\nOUT=" + headlabel

        viz.node(
            node_uuid,
            label=label,
            fontsize="10",
            style="filled",
            fillcolor=fillcolor,
        )

    for t in tsys.transitions:
        viz.edge(
            states_dictio[id(t.from_state)],
            states_dictio[id(t.to_state)],
            fontsize="8",
            label=str(t.name),
            taillabel=get_perc(
                len(t.from_state.data["outgoing_events"]),
                len(t.data["events"]),
            ),
            headlabel=get_perc(
                len(t.to_state.data["ingoing_events"]), len(t.data["events"])
            ),
        )

    viz.attr(overlap="false")

    viz.format = image_format.replace("html", "plain-ext")

    return viz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.statistics.attributes.log.select import *
from pm4py.statistics.attributes.log.get import *
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.obj import Trace
from pm4py.statistics.attributes.log.get import get_attribute_values
from pm4py.util import exec_utils
from pm4py.util import xes_constants as xes
from pm4py.util.constants import (
    PARAMETER_CONSTANT_ATTRIBUTE_KEY,
    PARAMETER_CONSTANT_ACTIVITY_KEY,
)
from pm4py.util.constants import (
    PARAMETER_CONSTANT_CASEID_KEY,
    PARAMETER_KEY_CASE_GLUE,
)
from pm4py.util.xes_constants import DEFAULT_NAME_KEY
from copy import copy
from typing import Optional, Dict, Any, Union, List
from pm4py.objects.log.obj import EventLog, EventStream


class Parameters(Enum):
    ATTRIBUTE_KEY = PARAMETER_CONSTANT_ATTRIBUTE_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    PARAMETER_KEY_CASE_GLUE = PARAMETER_KEY_CASE_GLUE
    DECREASING_FACTOR = "decreasingFactor"
    POSITIVE = "positive"
    STREAM_FILTER_KEY1 = "stream_filter_key1"
    STREAM_FILTER_VALUE1 = "stream_filter_value1"
    STREAM_FILTER_KEY2 = "stream_filter_key2"
    STREAM_FILTER_VALUE2 = "stream_filter_value2"
    KEEP_ONCE_PER_CASE = "keep_once_per_case"


def apply_numeric(
    log: EventLog,
    int1: float,
    int2: float,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Apply a filter on cases (numerical filter)

    Parameters
    --------------
    log
        Log
    int1
        Lower bound of the interval
    int2
        Upper bound of the interval
    parameters
        Possible parameters of the algorithm

    Returns
    --------------
    filtered_df
        Filtered dataframe
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    case_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, xes.DEFAULT_TRACEID_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )
    case_attribute_prefix = exec_utils.get_param_value(
        Parameters.PARAMETER_KEY_CASE_GLUE,
        parameters,
        constants.CASE_ATTRIBUTE_PREFIX,
    )
    # stream_filter_key is helpful to filter on cases containing an event with an attribute
    # in the specified value set, but such events shall have an activity in
    # particular.

    stream_filter_key1 = exec_utils.get_param_value(
        Parameters.STREAM_FILTER_KEY1, parameters, None
    )
    stream_filter_value1 = exec_utils.get_param_value(
        Parameters.STREAM_FILTER_VALUE1, parameters, None
    )
    stream_filter_key2 = exec_utils.get_param_value(
        Parameters.STREAM_FILTER_KEY2, parameters, None
    )
    stream_filter_value2 = exec_utils.get_param_value(
        Parameters.STREAM_FILTER_VALUE2, parameters, None
    )

    conversion_parameters = copy(parameters)
    conversion_parameters["deepcopy"] = False

    stream = log_converter.apply(
        log,
        variant=log_converter.TO_EVENT_STREAM,
        parameters=conversion_parameters,
    )
    if stream_filter_key1 is not None:
        stream = EventStream(
            list(
                filter(
                    lambda x: stream_filter_key1 in x
                    and x[stream_filter_key1] == stream_filter_value1,
                    stream,
                )
            ),
            attributes=log.attributes,
            extensions=log.extensions,
            classifiers=log.classifiers,
            omni_present=log.omni_present,
            properties=log.properties,
        )
    if stream_filter_key2 is not None:
        stream = EventStream(
            list(
                filter(
                    lambda x: stream_filter_key2 in x
                    and x[stream_filter_key2] == stream_filter_value2,
                    stream,
                )
            ),
            attributes=log.attributes,
            extensions=log.extensions,
            classifiers=log.classifiers,
            omni_present=log.omni_present,
            properties=log.properties,
        )

    if positive:
        stream = EventStream(
            list(
                filter(
                    lambda x: attribute_key in x
                    and int1 <= x[attribute_key] <= int2,
                    stream,
                )
            ),
            attributes=log.attributes,
            extensions=log.extensions,
            classifiers=log.classifiers,
            omni_present=log.omni_present,
            properties=log.properties,
        )
    else:
        stream = EventStream(
            list(
                filter(
                    lambda x: attribute_key in x
                    and (x[attribute_key] < int1 or x[attribute_key] > int2),
                    stream,
                )
            ),
            attributes=log.attributes,
            extensions=log.extensions,
            classifiers=log.classifiers,
            omni_present=log.omni_present,
            properties=log.properties,
        )

    all_cases_ids = set(x[case_attribute_prefix + case_key] for x in stream)

    filtered_log = EventLog(
        list(),
        attributes=log.attributes,
        extensions=log.extensions,
        classifiers=log.classifiers,
        omni_present=log.omni_present,
        properties=log.properties,
    )

    for case in log:
        if case.attributes[case_key] in all_cases_ids:
            filtered_log.append(case)

    return filtered_log


def apply_numeric_events(
    log: EventLog,
    int1: float,
    int2: float,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Apply a filter on events (numerical filter)

    Parameters
    --------------
    log
        Log
    int1
        Lower bound of the interval
    int2
        Upper bound of the interval
    parameters
        Possible parameters of the algorithm:
            Parameters.ATTRIBUTE_KEY => indicates which attribute to filter
            Parameters.POSITIVE => keep or remove traces with such events?

    Returns
    --------------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )

    conversion_parameters = copy(parameters)
    conversion_parameters["deepcopy"] = False
    stream = log_converter.apply(
        log,
        variant=log_converter.TO_EVENT_STREAM,
        parameters=conversion_parameters,
    )
    if exec_utils.get_param_value(Parameters.POSITIVE, parameters, True):
        stream = EventStream(
            list(
                filter(
                    lambda x: attribute_key in x
                    and int1 <= x[attribute_key] <= int2,
                    stream,
                )
            ),
            attributes=log.attributes,
            extensions=log.extensions,
            classifiers=log.classifiers,
            omni_present=log.omni_present,
            properties=log.properties,
        )
    else:
        stream = EventStream(
            list(
                filter(
                    lambda x: attribute_key in x
                    and (x[attribute_key] < int1 or x[attribute_key] > int2),
                    stream,
                )
            ),
            attributes=log.attributes,
            extensions=log.extensions,
            classifiers=log.classifiers,
            omni_present=log.omni_present,
            properties=log.properties,
        )
    filtered_log = log_converter.apply(
        stream,
        variant=log_converter.Variants.TO_EVENT_LOG,
        parameters=conversion_parameters,
    )

    return filtered_log


def apply_events(
    log: EventLog,
    values: List[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Filter log by keeping only events with an attribute value that belongs to the provided values list

    Parameters
    -----------
    log
        log
    values
        Allowed attributes
    parameters
        Parameters of the algorithm, including:
            Parameters.ACTIVITY_KEY -> Attribute identifying the activity in the log
            Parameters.POSITIVE -> Indicate if events should be kept/removed

    Returns
    -----------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    conversion_parameters = copy(parameters)
    conversion_parameters["deepcopy"] = False

    stream = log_converter.apply(
        log,
        variant=log_converter.TO_EVENT_STREAM,
        parameters=conversion_parameters,
    )
    if positive:
        stream = EventStream(
            list(filter(lambda x: x[attribute_key] in values, stream)),
            attributes=log.attributes,
            extensions=log.extensions,
            classifiers=log.classifiers,
            omni_present=log.omni_present,
            properties=log.properties,
        )
    else:
        stream = EventStream(
            list(filter(lambda x: x[attribute_key] not in values, stream)),
            attributes=log.attributes,
            extensions=log.extensions,
            classifiers=log.classifiers,
            omni_present=log.omni_present,
            properties=log.properties,
        )

    filtered_log = log_converter.apply(
        stream,
        variant=log_converter.Variants.TO_EVENT_LOG,
        parameters=conversion_parameters,
    )

    return filtered_log


def apply(
    log: EventLog,
    values: List[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Filter log by keeping only traces that has/has not events with an attribute value that belongs to the provided
    values list

    Parameters
    -----------
    log
        Trace log
    values
        Allowed attributes
    parameters
        Parameters of the algorithm, including:
            Parameters.ACTIVITY_KEY -> Attribute identifying the activity in the log
            Parameters.POSITIVE -> Indicate if events should be kept/removed

    Returns
    -----------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    filtered_log = EventLog(
        list(),
        attributes=log.attributes,
        extensions=log.extensions,
        classifiers=log.classifiers,
        omni_present=log.omni_present,
        properties=log.properties,
    )
    for trace in log:
        new_trace = Trace()

        found = False
        for j in range(len(trace)):
            if attribute_key in trace[j]:
                attribute_value = trace[j][attribute_key]
                if attribute_value in values:
                    found = True

        if (found and positive) or (not found and not positive):
            new_trace = trace
        else:
            for attr in trace.attributes:
                new_trace.attributes[attr] = trace.attributes[attr]

        if len(new_trace) > 0:
            filtered_log.append(new_trace)
    return filtered_log


def apply_trace_attribute(
    log: EventLog,
    values: List[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Filter a log on the trace attribute values

    Parameters
    --------------
    log
        Event log
    values
        Allowed/forbidden values
    parameters
        Parameters of the algorithm, including:
            - Parameters.ATTRIBUTE_KEY: the attribute at the trace level to filter
            - Parameters.POSITIVE: boolean (keep/discard values)

    Returns
    --------------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    filtered_log = EventLog(
        list(),
        attributes=log.attributes,
        extensions=log.extensions,
        classifiers=log.classifiers,
        omni_present=log.omni_present,
        properties=log.properties,
    )
    for trace in log:
        if positive:
            if (
                attribute_key in trace.attributes
                and trace.attributes[attribute_key] in values
            ):
                filtered_log.append(trace)
        else:
            if (
                not attribute_key in trace.attributes
                or not trace.attributes[attribute_key] in values
            ):
                filtered_log.append(trace)

    return filtered_log


def filter_log_on_max_no_activities(
    log: EventLog,
    max_no_activities: int = 25,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Filter a log on a maximum number of activities

    Parameters
    -------------
    log
        Log
    max_no_activities
        Maximum number of activities
    parameters
        Parameters of the algorithm

    Returns
    -------------
    filtered_log
        Filtered version of the event log
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    activity_key = (
        parameters[PARAMETER_CONSTANT_ACTIVITY_KEY]
        if PARAMETER_CONSTANT_ACTIVITY_KEY in parameters
        else DEFAULT_NAME_KEY
    )
    parameters[PARAMETER_CONSTANT_ATTRIBUTE_KEY] = activity_key
    all_activities = sorted(
        [(x, y) for x, y in get_attribute_values(log, activity_key).items()],
        key=lambda x: x[1],
        reverse=True,
    )
    activities = all_activities[: min(len(all_activities), max_no_activities)]
    activities = [x[0] for x in activities]

    if len(activities) < len(all_activities):
        log = apply_events(log, activities, parameters=parameters)
    return log


def filter_log_by_attributes_threshold(
    log,
    attributes,
    variants,
    vc,
    threshold,
    attribute_key=xes.DEFAULT_NAME_KEY,
):
    """
    Keep only attributes which number of occurrences is above the threshold (or they belong to the first variant)

    Parameters
    ----------
    log
        Log
    attributes
        Dictionary of attributes associated with their count
    variants
        (If specified) Dictionary with variant as the key and the list of traces as the value
    vc
        List of variant names along with their count
    threshold
        Cutting threshold (remove attributes which number of occurrences is below the threshold)
    attribute_key
        (If specified) Specify the activity key in the log (default concept:name)

    Returns
    ----------
    filtered_log
        Filtered log
    """
    filtered_log = EventLog(
        list(),
        attributes=log.attributes,
        extensions=log.extensions,
        classifiers=log.classifiers,
        omni_present=log.omni_present,
        properties=log.properties,
    )
    fva = [
        x[attribute_key] for x in variants[vc[0][0]][0] if attribute_key in x
    ]
    for trace in log:
        new_trace = Trace()
        for j in range(len(trace)):
            if attribute_key in trace[j]:
                attribute_value = trace[j][attribute_key]
                if attribute_value in attributes:
                    if (
                        attribute_value in fva
                        and attribute_key == xes.DEFAULT_NAME_KEY
                    ) or attributes[attribute_value] >= threshold:
                        new_trace.append(trace[j])
        if len(new_trace) > 0:
            for attr in trace.attributes:
                new_trace.attributes[attr] = trace.attributes[attr]
            filtered_log.append(new_trace)
    return filtered_log


def filter_log_relative_occurrence_event_attribute(
    log: EventLog,
    min_relative_stake: float,
    parameters: Optional[Dict[Any, Any]] = None,
) -> EventLog:
    """
    Filters the event log keeping only the events having an attribute value which occurs:
    - in at least the specified (min_relative_stake) percentage of events, when Parameters.KEEP_ONCE_PER_CASE = False
    - in at least the specified (min_relative_stake) percentage of cases, when Parameters.KEEP_ONCE_PER_CASE = True

    Parameters
    -------------------
    log
        Event log
    min_relative_stake
        Minimum percentage of cases (expressed as a number between 0 and 1) in which the attribute should occur.
    parameters
        Parameters of the algorithm, including:
        - Parameters.ATTRIBUTE_KEY => the attribute to use (default: concept:name)
        - Parameters.KEEP_ONCE_PER_CASE => decides the level of the filter to apply
        (if the filter should be applied on the cases, set it to True).

    Returns
    ------------------
    filtered_log
        Filtered event log
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, xes.DEFAULT_NAME_KEY
    )
    keep_once_per_case = exec_utils.get_param_value(
        Parameters.KEEP_ONCE_PER_CASE, parameters, True
    )

    parameters_cp = copy(parameters)

    activities_occurrences = get_attribute_values(
        log, attribute_key, parameters=parameters_cp
    )

    if keep_once_per_case:
        # filter on cases
        filtered_attributes = set(
            x
            for x, y in activities_occurrences.items()
            if y >= min_relative_stake * len(log)
        )
    else:
        # filter on events
        filtered_attributes = set(
            x
            for x, y in activities_occurrences.items()
            if y >= min_relative_stake * sum(len(x) for x in log)
        )

    return apply_events(log, filtered_attributes, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils, constants
from pm4py.objects.ocel.util import filtering_utils
from copy import copy
from typing import Dict, Any, Optional, Collection, Union
from pm4py.algo.filtering.common.timestamp.timestamp_common import (
    get_dt_from_string,
)
from pm4py.objects.ocel.obj import OCEL
import datetime


class Parameters(Enum):
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    POSITIVE = "positive"


def apply(
    ocel: OCEL,
    values: Collection[Any],
    parameters: Optional[Dict[Any, Any]] = None,
) -> OCEL:
    """
    Filters the object-centric event log on the provided event attributes values

    Parameters
    ----------------
    ocel
        Object-centric event log
    values
        Collection of values
    parameters
        Parameters of the algorithm, including:
        - Parameters.ATTRIBUTE_KEY => the attribute that should be filtered
        - Parameters.POSITIVE => decides if the values should be kept (positive=True) or removed (positive=False)

    Returns
    ----------------
    ocel
        Filtered object-centric event log
    """
    if parameters is None:
        parameters = {}

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, ocel.event_activity
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    ocel = copy(ocel)
    if positive:
        ocel.events = ocel.events[ocel.events[attribute_key].isin(values)]
    else:
        ocel.events = ocel.events[~ocel.events[attribute_key].isin(values)]

    return filtering_utils.propagate_event_filtering(
        ocel, parameters=parameters
    )


def apply_timestamp(
    ocel: OCEL,
    min_timest: Union[datetime.datetime, str],
    max_timest: Union[datetime.datetime, str],
    parameters: Optional[Dict[Any, Any]] = None,
) -> OCEL:
    """
    Filters the object-centric event log keeping events in the provided timestamp range

    Parameters
    -----------------
    ocel
        Object-centric event log
    min_timest
        Left extreme of the allowed timestamp interval (provided in the format: YYYY-mm-dd HH:MM:SS)
    max_timest
        Right extreme of the allowed timestamp interval (provided in the format: YYYY-mm-dd HH:MM:SS)
    parameters
        Parameters of the algorithm, including:
        - Parameters.TIMESTAMP_KEY => the attribute to use as timestamp

    Returns
    -----------------
    filtered_ocel
        Filtered object-centric event log
    """
    if parameters is None:
        parameters = {}

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, ocel.event_timestamp
    )
    min_timest = get_dt_from_string(min_timest)
    max_timest = get_dt_from_string(max_timest)

    ocel = copy(ocel)
    ocel.events = ocel.events[ocel.events[timestamp_key] >= min_timest]
    ocel.events = ocel.events[ocel.events[timestamp_key] <= max_timest]

    return filtering_utils.propagate_event_filtering(
        ocel, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
"""
This module is based on:
The minimal coverability graph for Petri nets
from Alain Finkel
"""

import numpy as np
from pm4py.util import nx_utils
from pm4py.algo.analysis.woflan.graphs import utility as helper
from copy import copy


def minimal_coverability_tree(net, initial_marking, original_net=None):
    """
    This method computes the minimal coverability tree. It is part of a method to obtain a minial coverability graph
    :param net: Petri Net
    :param initial_marking: Initial Marking of the Petri Net
    :param original_net: Petri Net without short-circuited transition
    :return: Minimal coverability tree
    """

    def check_if_marking_already_in_processed_nodes(n, processed_nodes):
        for node in processed_nodes:
            if np.array_equal(G.nodes[node]["marking"], G.nodes[n]["marking"]):
                return True
        return False

    def is_m_smaller_than_other(m, processed_nodes):
        for node in processed_nodes:
            if all(np.less_equal(m, G.nodes[node]["marking"])):
                return True
        return False

    def is_m_greater_than_other(m, processed_nodes):
        for node in processed_nodes:
            if all(np.greater_equal(m, G.nodes[node]["marking"])):
                return True
        return False

    def get_first_smaller_marking_on_path(n, m2):
        path = nx_utils.shortest_path(G, source=0, target=n)
        for node in path:
            if all(np.less_equal(G.nodes[node]["marking"], m2)):
                return node
        return None

    def remove_subtree(tree, n):
        bfs_tree = nx_utils.bfs_tree(tree, n)
        for edge in bfs_tree.edges:
            tree.remove_edge(edge[0], edge[1])
        for node in bfs_tree.nodes:
            if node != n:
                tree.remove_node(node)
        return tree

    G = nx_utils.MultiDiGraph()

    incidence_matrix = helper.compute_incidence_matrix(net)
    firing_dict = helper.split_incidence_matrix(incidence_matrix, net)
    req_dict = helper.compute_firing_requirement(net)

    initial_mark = helper.convert_marking(net, initial_marking, original_net)
    j = 0
    unprocessed_nodes = list()
    G.add_node(j, marking=initial_mark)
    unprocessed_nodes.append(j)
    j += 1

    processed_nodes = set()

    while len(unprocessed_nodes) > 0:
        n = unprocessed_nodes.pop()
        if check_if_marking_already_in_processed_nodes(n, processed_nodes):
            processed_nodes.add(n)
        elif is_m_smaller_than_other(G.nodes[n]["marking"], processed_nodes):
            predecessors = sorted(list(G.predecessors(n)))
            G.remove_edge(predecessors[0], n)
            G.remove_node(n)
        elif is_m_greater_than_other(G.nodes[n]["marking"], processed_nodes):
            m2 = G.nodes[n]["marking"].copy()
            ancestor_bool = False
            ancestors = sorted(list(nx_utils.ancestors(G, n)))
            for ancestor in ancestors:
                if is_m_greater_than_other(G.nodes[n]["marking"], [ancestor]):
                    i = 0
                    while i < len(G.nodes[n]["marking"]):
                        if (
                            G.nodes[ancestor]["marking"][i]
                            < G.nodes[n]["marking"][i]
                        ):
                            m2[i] = np.inf
                        i += 1
            n1 = None
            for ancestor in ancestors:
                if all(np.less_equal(G.nodes[ancestor]["marking"], m2)):
                    n1 = get_first_smaller_marking_on_path(n, m2)
                    break
            if n1 is not None:
                ancestor_bool = True
                G.nodes[n1]["marking"] = m2.copy()
                subtree = sorted(list(nx_utils.bfs_tree(G, n1)))
                for node in subtree:
                    if node in processed_nodes:
                        processed_nodes.remove(node)
                    if node in unprocessed_nodes:
                        del unprocessed_nodes[unprocessed_nodes.index(node)]
                G = remove_subtree(G, n1)
                if n1 not in unprocessed_nodes:
                    unprocessed_nodes.append(n1)
            processed_nodes_copy = copy(processed_nodes)
            for node in processed_nodes_copy:
                if node in G.nodes:
                    if all(np.less_equal(G.nodes[node]["marking"], m2)):
                        subtree = nx_utils.bfs_tree(G, node)
                        for node in subtree:
                            if node in processed_nodes:
                                processed_nodes.remove(node)
                            if node in unprocessed_nodes:
                                del unprocessed_nodes[
                                    unprocessed_nodes.index(node)
                                ]
                        remove_subtree(G, node)
                        G.remove_node(node)
            if not ancestor_bool:
                if n not in unprocessed_nodes:
                    unprocessed_nodes.append(n)
        else:
            enabled_markings = helper.enabled_markings(
                firing_dict, req_dict, G.nodes[n]["marking"]
            )
            for el in enabled_markings:
                G.add_node(j, marking=el[0])
                G.add_edge(n, j, transition=el[1])
                if j not in unprocessed_nodes:
                    unprocessed_nodes.append(j)
                j += 1
            processed_nodes.add(n)

    return (G, firing_dict, req_dict)


def apply(net, initial_marking, original_net=None):
    """
    Apply method from the "outside".
    :param net: Petri Net object
    :param initial_marking: Initial marking of the Petri Net object
    :param original_net: Petri Net object without short-circuited transition. For better usability, initial set to None
    :return: MultiDiGraph networkx object
    """

    def detect_same_labelled_nodes(G):
        same_labels = {}
        for node in G.nodes:
            if np.array2string(G.nodes[node]["marking"]) not in same_labels:
                same_labels[np.array2string(G.nodes[node]["marking"])] = [node]
            else:
                same_labels[np.array2string(G.nodes[node]["marking"])].append(
                    node
                )
        return same_labels

    def merge_nodes_of_same_label(G, same_labels):
        for marking in same_labels:
            if len(same_labels[marking]) > 1:
                origin = same_labels[marking][0]
                i = 1
                while i < len(same_labels[marking]):
                    G = nx_utils.contracted_nodes(
                        G, origin, same_labels[marking][i]
                    )
                    i += 1
        return G

    mct, firing_dict, req_dict = minimal_coverability_tree(
        net, initial_marking, original_net
    )
    mcg = merge_nodes_of_same_label(mct, detect_same_labelled_nodes(mct))

    to_remove_edges = []
    for edge in mcg.edges:
        reachable_markings = helper.enabled_markings(
            firing_dict, req_dict, mcg.nodes[edge[0]]["marking"]
        )
        not_reachable = True
        for el in reachable_markings:
            if np.array_equal(el[0], mcg.nodes[edge[1]]["marking"]):
                not_reachable = False
                break
        if not_reachable:
            to_remove_edges.append(edge)
    for edge in to_remove_edges:
        mcg.remove_edge(edge[0], edge[1])
    return mcg



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union

import importlib.util

from pm4py.util import xes_constants, constants, exec_utils, pandas_utils
from pm4py.algo.discovery.causal import algorithm as causal_discovery


class Outputs(Enum):
    DFG = "dfg"
    SEQUENCE = "sequence"
    PARALLEL = "parallel"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    ACTIVITIES = "activities"
    SKIPPABLE = "skippable"
    ACTIVITIES_ALWAYS_HAPPENING = "activities_always_happening"
    MIN_TRACE_LENGTH = "min_trace_length"
    TRACE = "trace"


class Parameters(Enum):
    SORT_REQUIRED = "sort_required"
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    INDEX_KEY = "index_key"


DEFAULT_SORT_REQUIRED = True
DEFAULT_INDEX_KEY = "@@index"


def apply(
    lf,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, Any]:
    if parameters is None:
        parameters = {}

    if importlib.util.find_spec("polars") is None:
        raise RuntimeError(
            "Polars LazyFrame provided but 'polars' package is not installed."
        )

    import polars as pl  # type: ignore[import-untyped]
    from pm4py.algo.discovery.dfg.adapters.polars import df_statistics

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    caseid_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY, parameters, None
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    sort_required = exec_utils.get_param_value(
        Parameters.SORT_REQUIRED, parameters, DEFAULT_SORT_REQUIRED
    )
    index_key = exec_utils.get_param_value(
        Parameters.INDEX_KEY, parameters, DEFAULT_INDEX_KEY
    )

    required_columns = {caseid_key, activity_key, timestamp_key}
    if start_timestamp_key is not None:
        required_columns.add(start_timestamp_key)

    if hasattr(lf, "collect_schema"):
        lf_columns = lf.collect_schema().names()
    else:
        lf_columns = lf.columns

    missing_columns = [col for col in required_columns if col not in lf_columns]
    if missing_columns:
        raise Exception(
            "The provided Polars LazyFrame does not contain the following required columns: "
            + ", ".join(sorted(missing_columns))
        )

    lf_selected = lf.select(list(required_columns))

    if sort_required:
        lf_selected = pandas_utils.insert_index(
            lf_selected, column_name=index_key, copy_dataframe=False
        )

        sort_columns = [caseid_key]
        if start_timestamp_key is not None:
            sort_columns.append(start_timestamp_key)
        sort_columns.append(timestamp_key)
        sort_columns.append(index_key)

        lf_selected = lf_selected.sort(sort_columns)

    dfg = df_statistics.get_dfg_graph(
        lf_selected,
        measure="frequency",
        activity_key=activity_key,
        case_id_glue=caseid_key,
        timestamp_key=timestamp_key,
        sort_caseid_required=False,
        sort_timestamp_along_case_id=False,
        start_timestamp_key=start_timestamp_key,
    )

    activities_df = (
        lf_selected.select(pl.col(activity_key).unique().alias("__act__"))
        .collect()
    )
    activities = set(activities_df["__act__"].to_list())

    grouped = lf_selected.group_by(caseid_key)
    first_df = grouped.agg(pl.col(activity_key).first().alias("__first__"))
    last_df = grouped.agg(pl.col(activity_key).last().alias("__last__"))

    start_activities = set(first_df.collect()["__first__"].to_list())
    end_activities = set(last_df.collect()["__last__"].to_list())

    parallel = {(x, y) for (x, y) in dfg if (y, x) in dfg}
    sequence = set(
        causal_discovery.apply(dfg, causal_discovery.Variants.CAUSAL_ALPHA)
    )

    min_trace_length_df = grouped.agg(pl.len().alias("__size__")).collect()
    min_trace_length = int(min_trace_length_df["__size__"].min())

    ret: Dict[str, Any] = {}
    ret[Outputs.DFG.value] = dfg
    ret[Outputs.SEQUENCE.value] = sequence
    ret[Outputs.PARALLEL.value] = parallel
    ret[Outputs.ACTIVITIES.value] = activities
    ret[Outputs.START_ACTIVITIES.value] = start_activities
    ret[Outputs.END_ACTIVITIES.value] = end_activities
    ret[Outputs.MIN_TRACE_LENGTH.value] = min_trace_length

    return ret




'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
def apply(net, im, fm, parameters=None):
    """
    Converts an accepting Petri net into a BPMN diagram

    Parameters
    --------------
    accepting_petri_net
        Accepting Petri net (list containing net + im + fm)
    parameters
        Parameters of the algorithm

    Returns
    --------------
    bpmn_graph
        BPMN diagram
    """
    if parameters is None:
        parameters = {}

    from pm4py.objects.bpmn.obj import BPMN
    from pm4py.objects.bpmn.util import reduction

    bpmn_graph = BPMN()

    entering_dictio = {}
    exiting_dictio = {}

    for place in net.places:
        node = BPMN.ExclusiveGateway()
        bpmn_graph.add_node(node)
        entering_dictio[place] = node
        exiting_dictio[place] = node

    for trans in net.transitions:
        if trans.label is None:
            if len(trans.in_arcs) > 1:
                node = BPMN.ParallelGateway(
                    gateway_direction=BPMN.Gateway.Direction.CONVERGING
                )
            elif len(trans.out_arcs) > 1:
                node = BPMN.ParallelGateway(
                    gateway_direction=BPMN.Gateway.Direction.DIVERGING
                )
            else:
                node = BPMN.ExclusiveGateway(
                    gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED
                )
            bpmn_graph.add_node(node)
            entering_dictio[trans] = node
            exiting_dictio[trans] = node
        else:
            if len(trans.in_arcs) > 1:
                entering_node = BPMN.ParallelGateway(
                    gateway_direction=BPMN.Gateway.Direction.CONVERGING
                )
            else:
                entering_node = BPMN.ExclusiveGateway(
                    gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED
                )

            if len(trans.out_arcs) > 1:
                exiting_node = BPMN.ParallelGateway(
                    gateway_direction=BPMN.Gateway.Direction.DIVERGING
                )
            else:
                exiting_node = BPMN.ExclusiveGateway(
                    gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED
                )

            task = BPMN.Task(name=trans.label)
            bpmn_graph.add_node(task)

            bpmn_graph.add_flow(BPMN.SequenceFlow(entering_node, task))
            bpmn_graph.add_flow(BPMN.SequenceFlow(task, exiting_node))

            entering_dictio[trans] = entering_node
            exiting_dictio[trans] = exiting_node

    for arc in net.arcs:
        bpmn_graph.add_flow(
            BPMN.SequenceFlow(
                exiting_dictio[arc.source], entering_dictio[arc.target]
            )
        )

    start_node = BPMN.StartEvent(name="start", isInterrupting=True)
    end_node = BPMN.NormalEndEvent(name="end")
    bpmn_graph.add_node(start_node)
    bpmn_graph.add_node(end_node)
    for place in im:
        bpmn_graph.add_flow(
            BPMN.SequenceFlow(start_node, entering_dictio[place])
        )
    for place in fm:
        bpmn_graph.add_flow(BPMN.SequenceFlow(exiting_dictio[place], end_node))

    bpmn_graph = reduction.apply(bpmn_graph)

    for node in bpmn_graph.get_nodes():
        node.set_process(bpmn_graph.get_process_id())

    for edge in bpmn_graph.get_flows():
        edge.set_process(bpmn_graph.get_process_id())

    return bpmn_graph



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.petri_net.saw_net.obj import StochasticArcWeightNet
from pm4py.objects.petri_net.stochastic.obj import StochasticPetriNet
from pm4py.objects.petri_net.saw_net.semantics import (
    GlobalStochasticArcWeightNetSemantics as sawsem_global,
    LocalStochasticArcWeightNetSemantics as sawsem_local,
)
from typing import Counter as TCounter, Tuple
from collections import deque


def convert_saw_net_to_stochastic_net_global_semantics(
    saw: StochasticArcWeightNet,
) -> StochasticPetriNet:
    sn = StochasticPetriNet()
    p_map = dict()
    for p in saw.places:
        p_new = StochasticPetriNet.Place(p.name, properties=p.properties)
        sn.places.add(p_new)
        p_map[p] = p_new

    a_sums = dict()
    for a in saw.arcs:
        a_sums[a] = 0
        for w, v in a.weight_distribution.items():
            a_sums[a] = a_sums[a] + v

    for t in saw.transitions:
        bindings = sawsem_global.all_legal_bindings(saw, t)
        for b in bindings:
            transition_weight = t.weight
            t_new = StochasticPetriNet.Transition(
                "t" + str(len(sn.transitions)), t.label
            )
            sn.transitions.add(t_new)
            for a, w in b:
                a_new = (
                    StochasticPetriNet.Arc(t_new, p_map[a.target], w)
                    if a.source == t
                    else StochasticPetriNet.Arc(p_map[a.source], t_new, w)
                )
                if a.source == t:
                    t_new.out_arcs.add(a_new)
                    p_map[a.target].in_arcs.add(a_new)
                else:
                    t_new.in_arcs.add(a_new)
                    p_map[a.source].out_arcs.add(a_new)
                sn.arcs.add(a_new)
                transition_weight = transition_weight * (
                    a.weight_distribution[w] / a_sums[a]
                )
            t_new.weight = transition_weight
    return sn, p_map


def convert_saw_net_to_stochastic_net_local_semantics(
    saw: StochasticArcWeightNet,
    marking: TCounter[StochasticArcWeightNet.Place],
) -> Tuple[StochasticPetriNet, TCounter[StochasticArcWeightNet.Place]]:
    sn = StochasticPetriNet()
    open = deque()
    closed = set()
    markings = dict()
    markings[marking] = {
        p: StochasticPetriNet.Place(p.name + "-" + str(marking))
        for p in marking
        if marking[p] > 0
    }
    for p in markings[marking]:
        sn.places.add(markings[marking][p])
    open.append(marking)
    while len(open) > 0:
        m = open.pop()
        for t in saw.transitions:
            if sawsem_local.is_enabled(saw, t, m):
                for b in sawsem_local.all_enabled_bindings(saw, t, m):
                    m_out = sawsem_local.fire(saw, b, m)
                    if m_out not in markings:
                        markings[m_out] = {
                            p: StochasticPetriNet.Place(
                                p.name + "-" + str(m_out)
                            )
                            for p in m_out
                            if m_out[p] > 0
                        }
                        for p in markings[m_out]:
                            sn.places.add(markings[m_out][p])
                    t_new = StochasticPetriNet.Transition(
                        t.name + "-" + str(b),
                        t.label,
                        weight=t.weight
                        * sawsem_local.amortized_priority(b)
                        / sum(
                            [
                                sawsem_local.amortized_priority(bb)
                                for bb in sawsem_local.all_enabled_bindings(
                                    saw, t, m
                                )
                            ]
                        ),
                    )
                    sn.transitions.add(t_new)
                    plcs_in = markings[m]
                    plcs_out = markings[m_out]
                    for p in m:
                        if m[p] > 0:
                            a = StochasticPetriNet.Arc(plcs_in[p], t_new, m[p])
                            plcs_in[p].out_arcs.add(a)
                            t_new.in_arcs.add(a)
                            sn.arcs.add(a)
                    for p in m_out:
                        if m_out[p] > 0:
                            a = StochasticPetriNet.Arc(
                                t_new, plcs_out[p], m_out[p]
                            )
                            plcs_out[p].in_arcs.add(a)
                            t_new.out_arcs.add(a)
                            sn.arcs.add(a)
                    if m_out not in closed and m_out is not m:
                        open.append(m_out)
    return sn, {markings[marking][p]: marking[p] for p in marking}



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.util import constants
from pm4py.util import exec_utils, pandas_utils
from pm4py.util import xes_constants as xes
from pm4py.util.constants import CASE_CONCEPT_NAME
from typing import Optional, Dict, Any, Union, List
import pandas as pd


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    PARAMETER_SAMPLE_SIZE = "sample_size"
    SORT_LOG_REQUIRED = "sort_log_required"


def apply(
    dataframe: pd.DataFrame,
    list_activities: List[str],
    sample_size: int,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, Any]:
    """
    Finds the performance spectrum provided a dataframe
    and a list of activities

    Parameters
    -------------
    dataframe
        Dataframe
    list_activities
        List of activities interesting for the performance spectrum (at least two)
    sample_size
        Size of the sample
    parameters
        Parameters of the algorithm,  including:
            - Parameters.ACTIVITY_KEY
            - Parameters.TIMESTAMP_KEY
            - Parameters.CASE_ID_KEY

    Returns
    -------------
    points
        Points of the performance spectrum
    """
    if parameters is None:
        parameters = {}

    import pandas as pd
    import numpy as np

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, xes.DEFAULT_TIMESTAMP_KEY
    )
    sort_log_required = exec_utils.get_param_value(
        Parameters.SORT_LOG_REQUIRED, parameters, True
    )

    dataframe = dataframe[[case_id_glue, activity_key, timestamp_key]]
    dataframe[activity_key] = dataframe[activity_key].astype("string")
    dataframe = dataframe[dataframe[activity_key].isin(list_activities)]
    dataframe = pandas_utils.insert_index(
        dataframe, constants.DEFAULT_EVENT_INDEX_KEY
    )
    if sort_log_required:
        dataframe = dataframe.sort_values(
            [case_id_glue, timestamp_key, constants.DEFAULT_EVENT_INDEX_KEY]
        )
    dataframe[timestamp_key] = (
        dataframe[timestamp_key].astype(np.int64) / 10**9
    )

    def key(k, n):
        return k + str(n)

    # create a dataframe with all needed columns to check for the activities
    # pattern
    dfs = [
        dataframe.add_suffix(str(i)).shift(-i)
        for i in range(len(list_activities))
    ]
    dataframe = pandas_utils.concat(dfs, axis=1)
    # keep only rows that belong to exactly one case
    for i in range(len(list_activities) - 1):
        dataframe = dataframe[
            dataframe[key(case_id_glue, i)]
            == dataframe[key(case_id_glue, i + 1)]
        ]

    column_list = [key(activity_key, i) for i in range(len(list_activities))]
    pattern = "".join(list_activities)
    # keep only rows that have the desired activities pattern

    matches = dataframe[
        np.equal(dataframe[column_list].agg("".join, axis=1), pattern)
    ]
    if len(matches) > sample_size:
        matches = matches.sample(n=sample_size)

    filt_col_names = [
        timestamp_key + str(i) for i in range(len(list_activities))
    ]
    points = pandas_utils.to_dict_records(matches)
    points = [[p[tk] for tk in filt_col_names] for p in points]
    points = sorted(points, key=lambda x: x[0])

    return points



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from enum import Enum
from pm4py.util import exec_utils


class Parameters(Enum):
    EVENT_NUM_ATTRIBUTES = "num_ev_attr"


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Enables the extraction of a given collection of numeric event attributes in the feature table
    (specified inside the "num_ev_attr" parameter).

    Parameters
    ----------------
    ocel
        OCEL
    parameters
        Parameters of the algorithm:
            - Parameters.EVENT_NUM_ATTRIBUTES => collection of numeric attributes to consider for feature extraction

    Returns
    ----------------
    data
        Extracted feature values
    feature_names
        Feature names
    """
    if parameters is None:
        parameters = {}

    ordered_events = (
        parameters["ordered_events"]
        if "ordered_events" in parameters
        else ocel.events[ocel.event_id_column].to_numpy()
    )

    data = []
    feature_names = []

    event_num_attributes = exec_utils.get_param_value(
        Parameters.EVENT_NUM_ATTRIBUTES, parameters, None
    )

    if event_num_attributes:
        feature_names = feature_names + [
            "@@event_num_" + x for x in event_num_attributes
        ]

        attr_values = {}
        for attr in event_num_attributes:
            values = (
                ocel.events[[ocel.event_id_column, attr]]
                .dropna(subset=[attr])
                .to_dict("records")
            )
            values = {x[ocel.event_id_column]: x[attr] for x in values}
            attr_values[attr] = values

        for ev in ordered_events:
            data.append([])
            for attr in event_num_attributes:
                data[-1].append(
                    float(attr_values[attr][ev])
                    if ev in attr_values[attr]
                    else 0.0
                )

    return data, feature_names



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
from typing import Dict, Collection, Any, Tuple

from pm4py.util import nx_utils

from pm4py.objects.dfg.obj import DFG


def get_vertices(dfg: DFG) -> Collection[Any]:
    """
    Returns the vertices of the dfg

    :param dfg: input directly follows graph

    :rtype: ``Collection[Any]``
    """
    alphabet = set()
    [alphabet.update({a, b}) for (a, b) in dfg.graph]
    [alphabet.update({a}) for a in dfg.start_activities]
    [alphabet.update({a}) for a in dfg.end_activities]
    return alphabet


def get_outgoing_arcs(dfg: DFG) -> Dict[Any, Dict[Any, int]]:
    """
    Returns the outgoing arcs of the provided DFG graph.
    Returns a dictionary mapping each 'source' node onto its set of 'target' nodes and associated frequency.

    :param dfg: ``DFG`` object

    :rtype: ``Dict[str, Counter[str]]``

    """
    outgoing = {a: Counter() for a in get_vertices(dfg)}
    for a, b in dfg.graph:
        outgoing[a][b] = (
            dfg.graph[(a, b)]
            if b not in outgoing[a]
            else outgoing[a][b] + dfg.graph[(a, b)]
        )
    return outgoing


def get_incoming_arcs(dfg: DFG) -> Dict[Any, Dict[Any, int]]:
    """
    Returns the incoming arcs of the provided DFG graph.
    Returns a dictionary mapping each 'target' node onto its set of 'source' nodes and associated frequency.

    :param dfg: ``DFG`` object

    :rtype: ``Dict[str, Counter[str]]``

    """
    incoming = {a: Counter() for a in get_vertices(dfg)}
    for a, b in dfg.graph:
        incoming[b][a] = (
            dfg.graph[(a, b)]
            if a not in incoming[b]
            else incoming[b][a] + dfg.graph[(a, b)]
        )
    return incoming


def get_source_vertices(dfg: DFG) -> Collection[Any]:
    """
    Gets source vertices from a Directly-Follows Graph.
    Vertices are returned that have no incoming arcs

    :param dfg: ``DFG`` object

    :rtype: ``Collection[Any]``
    """
    starters = set()
    incoming = get_incoming_arcs(dfg)
    [starters.add(a) for a in incoming if len(incoming[a]) == 0]
    return starters


def get_sink_vertices(dfg: DFG) -> Collection[Any]:
    """
    Gets sink vertices from a Directly-Follows Graph.
    Vertices are returned that have no outgoing arcs

    :param dfg: ``DFG`` object

    :rtype: ``Collection[Any]``
    """
    ends = set()
    outgoing = get_outgoing_arcs(dfg)
    [ends.add(a) for a in outgoing if len(outgoing[a]) == 0]
    return ends


def get_transitive_relations(
    dfg: DFG,
) -> Tuple[Dict[Any, Collection[Any]], Dict[Any, Collection[Any]]]:
    """
    Computes the full transitive relations in both directions (all activities reachable from a given activity and all
    activities that can reach the activity)

    :param dfg: ``DFG`` object

    :rtype: ``Tuple[Dict[Any, Collection[Any]], Dict[Any, Collection[Any]]] first argument maps an activity on all other
    activities that are able to reach the activity ('transitive pre set')
        second argument maps an activity on all other activities that it can reach (transitively) ('transitive post set')
    """
    G = nx_utils.DiGraph()
    alph = get_vertices(dfg)

    for act in alph:
        G.add_node(act)

    for edge in dfg.graph:
        G.add_edge(edge[0], edge[1])

    pre = {}
    post = {}

    for a in alph:
        pre[a] = nx_utils.ancestors(G, a)
        post[a] = nx_utils.descendants(G, a)

    return pre, post


def get_vertex_frequencies(dfg: DFG) -> Dict[Any, int]:
    """
    Computes the number of times a vertex in the dfg is visited.
    The number equals the number of occurrences in the underlying log and is computed by summing up the incoming
    arc frequency and the number of starts in the vertex. The value is equal to the number of outgoing arcs combined
    with the number of endings of the vertex.
    """
    c = Counter()
    for v in get_vertices(dfg):
        c[v] = 0
    for a, b in dfg.graph:
        c[a] += dfg.graph[(a, b)]
    for a in dfg.start_activities:
        c[a] += dfg.start_activities[a]
    return c


def as_nx_graph(dfg: DFG):
    nx_graph = nx_utils.DiGraph()
    nx_graph.add_nodes_from(get_vertices(dfg))
    for a, b in dfg.graph:
        nx_graph.add_edge(a, b)
    return nx_graph


def get_edges(dfg: DFG) -> Collection[Tuple[Any, Any]]:
    return dfg.graph.keys()



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter


def equivalence(trace):
    """
    Get the equivalence relations given a list of activities

    Parameters
    --------------
    trace
        List activities

    Returns
    --------------
    rel
        Relations inside the trace
    """
    ret = list()
    freq = activ_freq(trace)
    for x in freq:
        for y in freq:
            if x != y and freq[x] == freq[y]:
                for i in range(freq[x]):
                    ret.append((x, y))
    return ret


def after(trace):
    """
    Get the after- relations given a list of activities

    Parameters
    --------------
    trace
        List activities

    Returns
    --------------
    rel
        After- inside the trace
    """
    return list(
        (trace[i], trace[j])
        for i in range(len(trace))
        for j in range(len(trace))
        if j > i
    )


def before(trace):
    """
    Get the before- relations given a list of activities

    Parameters
    --------------
    trace
        List activities

    Returns
    --------------
    rel
        Before- inside the trace
    """
    return list(
        (trace[i], trace[j])
        for i in range(len(trace))
        for j in range(len(trace))
        if j < i
    )


def combos(trace):
    """
    Get the combinations between all the activities of the trace relations given a list of activities

    Parameters
    --------------
    trace
        List activities

    Returns
    --------------
    rel
        Combos inside the trace
    """
    return set((x, y) for x in trace for y in trace if x != y)


def directly_follows(trace):
    """
    Get the directly-follows relations given a list of activities

    Parameters
    --------------
    trace
        List activities

    Returns
    --------------
    rel
        Directly-follows relations inside the trace
    """
    return list((trace[i], trace[i + 1]) for i in range(len(trace) - 1))


def activ_freq(trace):
    """
    Gets the frequency of activities happening in a trace

    Parameters
    --------------
    trace
        List activities

    Returns
    --------------
    freq
        Frequency of activities
    """
    return Counter(trace)


def get_trace_info(trace):
    """
    Technical method for conformance checking
    """
    return (
        equivalence(trace),
        after(trace),
        before(trace),
        combos(trace),
        directly_follows(trace),
        activ_freq(trace),
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import PARAMETER_CONSTANT_ATTRIBUTE_KEY
from pm4py.util.xes_constants import DEFAULT_NAME_KEY
from pm4py.objects.log.obj import EventLog, Trace, EventStream
from pm4py.objects.conversion.log import converter as log_converter
from enum import Enum
from pm4py.util import exec_utils


class Parameters(Enum):
    ATTRIBUTE_KEY = PARAMETER_CONSTANT_ATTRIBUTE_KEY
    POSITIVE = "positive"


def filter_log_events_attr(log, values, parameters=None):
    """
    Filter log by keeping only events with an attribute value that belongs to the provided values list

    Parameters
    -----------
    log
        log
    values
        Allowed attributes
    parameters
        Parameters of the algorithm, including:
            activity_key -> Attribute identifying the activity in the log
            positive -> Indicate if events should be kept/removed

    Returns
    -----------
    filtered_log
        Filtered log
    """

    # CODE SAVING FROM FILTERS

    if parameters is None:
        parameters = {}

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    stream = log_converter.apply(
        log,
        variant=log_converter.TO_EVENT_STREAM,
        parameters={"deepcopy": False},
    )
    if positive:
        stream = EventStream(
            list(filter(lambda x: x[attribute_key] in values, stream))
        )
    else:
        stream = EventStream(
            list(filter(lambda x: x[attribute_key] not in values, stream))
        )

    filtered_log = log_converter.apply(
        stream, variant=log_converter.Variants.TO_EVENT_LOG
    )

    return filtered_log


def filter_log_traces_attr(log, values, parameters=None):
    """
    Filter log by keeping only traces that has/has not events with an attribute value that belongs to the provided
    values list

    Parameters
    -----------
    log
        Trace log
    values
        Allowed attributes
    parameters
        Parameters of the algorithm, including:
            activity_key -> Attribute identifying the activity in the log
            positive -> Indicate if events should be kept/removed

    Returns
    -----------
    filtered_log
        Filtered log
    """

    # CODE SAVING FROM FILTERS

    if parameters is None:
        parameters = {}

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    filtered_log = EventLog()
    for trace in log:
        new_trace = Trace()

        found = False
        for j in range(len(trace)):
            if attribute_key in trace[j]:
                attribute_value = trace[j][attribute_key]
                if attribute_value in values:
                    found = True

        if (found and positive) or (not found and not positive):
            new_trace = trace
        else:
            for attr in trace.attributes:
                new_trace.attributes[attr] = trace.attributes[attr]

        if len(new_trace) > 0:
            filtered_log.append(new_trace)
    return filtered_log



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np


def get_log_encoded(
    event_log, trace_attributes=[], event_attributes=[], concatenate=False
):
    """
    Get event log encoded into matrix.

    Parameters
    ------------
    event_log
        Trace log
    trace_attributes
        Attributes of the trace to be encoded
    event_attributes
        Attributes of the events to be encoded
    concatenate
        Boolean indicating if to generate all sub-sequences of events in a trace

    Returns
    ------------
    dataset
        A numpy matrix with the event log
    columns
        The names of the columns in the dataset
    """
    columns = []
    dataset = []

    max_trace_len = 0

    for trace_index, trace in enumerate(event_log):
        trace_encoding = []
        tr_columns = []

        for trace_attribute in trace_attributes:
            tr_columns.append(trace_attribute)

            try:
                attr = trace.attributes[trace_attribute]
            except BaseException:
                attr = None
            trace_encoding.append(attr)

            for event_index, event in enumerate(trace):
                for event_attribute in event_attributes:
                    tr_columns.append(event_attribute)

                    try:
                        attr = event[event_attribute]
                    except BaseException:
                        attr = None
                    trace_encoding.append(attr)

                # For each trace in the event log, sequentially append the
                # event sequence until that event
                if concatenate is True:
                    if len(trace_encoding) > max_trace_len:
                        max_trace_len = len(trace_encoding)
                        columns = tr_columns

                    dataset.append(np.asarray(trace_encoding))

            if concatenate is not True:
                if len(trace_encoding) > max_trace_len:
                    max_trace_len = len(trace_encoding)
                    columns = tr_columns

                dataset.append(np.asarray(trace_encoding))

    dataset = np.asarray(
        [
            np.pad(
                a, (0, max_trace_len - len(a)), "constant", constant_values=0
            )
            for a in dataset
        ]
    )

    columns = np.asarray(columns)

    return dataset, columns



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union

import importlib.util

from pm4py.util import exec_utils, constants, xes_constants, pandas_utils


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY


DIFF_INDEX = "@@diff_index"
NEXT_INDEX = "@@next_index"


def apply(df, parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> Dict[str, int]:
    if parameters is None:
        parameters = {}

    if importlib.util.find_spec("polars") is None:
        raise RuntimeError(
            "Polars LazyFrame provided but 'polars' package is not installed."
        )

    import polars as pl  # type: ignore[import-untyped]

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    required_columns = {activity_key, case_id_key}
    if hasattr(df, "collect_schema"):
        df_columns = df.collect_schema().names()
    else:
        df_columns = df.columns
    missing_columns = [col for col in required_columns if col not in df_columns]
    if missing_columns:
        raise Exception(
            "The provided Polars LazyFrame does not contain the following required columns: "
            + ", ".join(sorted(missing_columns))
        )

    lf = df.select(list(required_columns))
    lf = pandas_utils.insert_ev_in_tr_index(
        lf,
        case_id=case_id_key,
        column_name=constants.DEFAULT_INDEX_IN_TRACE_KEY,
    )

    lf = lf.sort([case_id_key, constants.DEFAULT_INDEX_IN_TRACE_KEY])

    lf = lf.with_columns(
        pl.col(constants.DEFAULT_INDEX_IN_TRACE_KEY)
        .shift(-1)
        .over([case_id_key, activity_key])
        .alias(NEXT_INDEX)
    )

    lf = lf.with_columns(
        (pl.col(NEXT_INDEX) - pl.col(constants.DEFAULT_INDEX_IN_TRACE_KEY) - 1).alias(
            DIFF_INDEX
        )
    )

    lf = lf.filter(pl.col(DIFF_INDEX).is_not_null() & (pl.col(DIFF_INDEX) >= 0))

    result = (
        lf.group_by(activity_key)
        .agg(pl.col(DIFF_INDEX).min().alias(DIFF_INDEX))
        .collect()
    )

    return {
        row[activity_key]: int(row[DIFF_INDEX])
        for row in result.iter_rows(named=True)
    }



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union, Tuple, List

import numpy as np

from pm4py.objects.log.obj import EventLog, Trace
from pm4py.util import constants
from pm4py.util import exec_utils, variants_util
from pm4py.util.xes_constants import DEFAULT_TIMESTAMP_KEY
from pm4py.objects.conversion.log import converter as log_converter


class Parameters(Enum):
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    MAX_NO_POINTS_SAMPLE = "max_no_of_points_to_sample"
    KEEP_ONCE_PER_CASE = "keep_once_per_case"


def get_language(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Union[Dict[List[str], float], Dict[str, float]]:
    """
    Gets the stochastic language of the log (from the variants)

    Parameters
    --------------
    log
        Event log
    parameters
        Parameters

    Returns
    --------------
    dictio
        Dictionary containing the stochastic language of the log
        (variant associated to a number between 0 and 1; the sum is 1)
    """
    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )
    vars = get_variants(log, parameters=parameters)
    vars = {
        variants_util.get_activities_from_variant(x): len(y)
        for x, y in vars.items()
    }

    all_values_sum = sum(vars.values())
    for x in vars:
        vars[x] = vars[x] / all_values_sum
    return vars


def get_variants(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Union[Dict[List[str], List[Trace]], Dict[str, List[Trace]]]:
    """
    Gets a dictionary whose key is the variant and as value there
    is the list of traces that share the variant

    Parameters
    ----------
    log
        Trace log
    parameters
        Parameters of the algorithm, including:
            Parameters.ACTIVITY_KEY -> Attribute identifying the activity in the log

    Returns
    ----------
    variant
        Dictionary with variant as the key and the list of traces as the value
    """
    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    variants_trace_idx = get_variants_from_log_trace_idx(
        log, parameters=parameters
    )

    all_var = convert_variants_trace_idx_to_trace_obj(log, variants_trace_idx)

    return all_var


def get_variants_along_with_case_durations(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[
    Union[Dict[List[str], List[Trace]], Dict[str, List[Trace]]], np.array
]:
    """
    Gets a dictionary whose key is the variant and as value there
    is the list of traces that share the variant

    Parameters
    ----------
    log
        Trace log
    parameters
        Parameters of the algorithm, including:
            Parameters.ACTIVITY_KEY -> Attribute identifying the activity in the log

    Returns
    ----------
    variant
        Dictionary with variant as the key and the list of traces as the value
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )

    variants_trace_idx = get_variants_from_log_trace_idx(
        log, parameters=parameters
    )
    all_var = convert_variants_trace_idx_to_trace_obj(log, variants_trace_idx)

    all_durations = {}

    for var in all_var:
        all_durations[var] = []
        for trace in all_var[var]:
            if (
                trace
                and timestamp_key in trace[-1]
                and timestamp_key in trace[0]
            ):
                all_durations[var].append(
                    (
                        trace[-1][timestamp_key] - trace[0][timestamp_key]
                    ).total_seconds()
                )
            else:
                all_durations[var].append(0)
        all_durations[var] = np.array(all_durations[var])

    return all_var, all_durations


def get_variants_from_log_trace_idx(log, parameters=None):
    """
    Gets a dictionary whose key is the variant and as value there
    is the list of traces indexes that share the variant

    Parameters
    ----------
    log
        Log
    parameters
        Parameters of the algorithm, including:
            Parameters.ACTIVITY_KEY -> Attribute identifying the activity in the log

    Returns
    ----------
    variant
        Dictionary with variant as the key and the list of traces indexes as the value
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    variants = {}
    for trace_idx, trace in enumerate(log):
        variant = variants_util.get_variant_from_trace(
            trace, parameters=parameters
        )
        if variant not in variants:
            variants[variant] = []
        variants[variant].append(trace_idx)

    return variants


def get_variants_sorted_by_count(variants):
    """
    From the dictionary of variants returns an ordered list of variants
    along with their count

    Parameters
    ----------
    variants
        Dictionary with variant as the key and the list of traces as the value

    Returns
    ----------
    var_count
        List of variant names along with their count
    """
    var_count = []
    for variant in variants:
        var_count.append([variant, len(variants[variant])])
    var_count = sorted(var_count, key=lambda x: (x[1], x[0]), reverse=True)
    return var_count


def convert_variants_trace_idx_to_trace_obj(log, variants_trace_idx):
    """
    Converts variants expressed as trace indexes to trace objects

    Parameters
    -----------
    log
        Trace log object
    variants_trace_idx
        Variants associated to a list of belonging indexes

    Returns
    -----------
    variants
        Variants associated to a list of belonging traces
    """
    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG)

    variants = {}

    for key in variants_trace_idx:
        variants[key] = []
        for value in variants_trace_idx[key]:
            variants[key].append(log[value])

    return variants



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.statistics.passed_time.polars.variants import pre, post
from typing import Optional, Dict, Any
import polars as pl


def apply(
    lf: pl.LazyFrame,
    activity: str,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, Any]:
    """
    Gets the time passed from/to each preceding/succeeding activity

    Parameters
    -------------
    lf
        LazyFrame
    activity
        Activity that we are considering
    parameters
        Possible parameters of the algorithm

    Returns
    -------------
    dictio
        Dictionary containing both 'pre' and 'post' keys with the
        list of aggregates times from/to each preceding/succeeding activity
    """
    if parameters is None:
        parameters = {}

    # Get pre statistics
    pre_stats = pre.apply(lf, activity, parameters=parameters)
    
    # Get post statistics  
    post_stats = post.apply(lf, activity, parameters=parameters)
    
    # Combine results
    result = {}
    result.update(pre_stats)
    result.update(post_stats)
    
    return result


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.algo.transformation.log_to_target.variants import (
    next_activity,
    next_time,
    remaining_time,
)
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
from typing import Union, Dict, Optional, Any, Tuple, List
from pm4py.util import exec_utils


class Variants(Enum):
    NEXT_ACTIVITY = next_activity
    NEXT_TIME = next_time
    REMAINING_TIME = remaining_time


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    variant=None,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Extracts from the event log
    the target vector for a specific ML use case

    Parameters
    ---------------
    log
        Event log / Event stream / Pandas dataframe
    variant
        Specification of the target vector:
        - Variants.NEXT_ACTIVITY => encodes the next activity
        - Variants.NEXT_TIME => encodes the next timestamp
        - Variants.REMAINING_TIME => encodes the remaining time

    Returns
    --------------
    vector
        Target vector for the specified ML use case
    classes
        Classes (for every column of the target vector)
    """
    if variant is None:
        raise Exception(
            "please provide the variant between: Variants.NEXT_ACTIVITY, Variants.NEXT_TIME, Variants.REMAINING_TIME"
        )
    return exec_utils.get_variant(variant).apply(log, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import pandas as pd
from enum import Enum
from pm4py.util import constants, xes_constants, pandas_utils, exec_utils
import numpy as np
from collections import Counter
from typing import Tuple, Dict, Collection
import importlib.util


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    INDEX_KEY = "index_key"


def apply(
    dataframe: pd.DataFrame, parameters=None
) -> Tuple[Dict[Collection[str], int], Dict[str, Collection[str]]]:
    """
    Efficient method returning the variants from a Pandas dataframe (through Numpy)

    Minimum viable example:

        import pandas as pd
        import pm4py
        from pm4py.objects.log.util import pandas_numpy_variants

        dataframe = pd.read_csv('tests/input_data/receipt.csv')
        dataframe = pm4py.format_dataframe(dataframe)
        variants_dict, case_variant = pandas_numpy_variants.apply(dataframe)


    Parameters
    ------------------
    dataframe
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY => the case identifier
        - Parameters.ACTIVITY_KEY => the activity
        - Parameters.TIMESTAMP_KEY => the timestamp
        - Parameters.INDEX_KEY => the index

    Returns
    ------------------
    variants_dict
        Dictionary associating to each variant the number of occurrences in the dataframe
    case_variant
        Dictionary associating to each case identifier the corresponding variant
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    index_key = exec_utils.get_param_value(
        Parameters.INDEX_KEY, parameters, constants.DEFAULT_INDEX_KEY
    )

    if not (hasattr(dataframe, "attrs") and dataframe.attrs):
        # dataframe has not been initialized through format_dataframe
        dataframe = pandas_utils.insert_index(dataframe, index_key)
        dataframe.sort_values([case_id_key, timestamp_key, index_key])

    case_variant = dict()

    if importlib.util.find_spec("cudf"):
        case_variant = (
            dataframe.groupby(case_id_key)[activity_key].agg(list).to_dict()
        )
        case_variant = {x: tuple(y) for x, y in case_variant.items()}
        variants_counter = Counter(case_variant.values())
    else:
        variants_counter = Counter()
        cases = dataframe[case_id_key].to_numpy()
        activities = dataframe[activity_key].to_numpy()

        c_unq, c_ind, c_counts = np.unique(
            cases, return_index=True, return_counts=True
        )

        for i in range(len(c_ind)):
            si = c_ind[i]
            ei = si + c_counts[i]
            acts = tuple(activities[si:ei])
            variants_counter[acts] += 1
            case_variant[c_unq[i]] = acts

    # return as Python dictionary
    variants_dict = {x: y for x, y in variants_counter.items()}

    return variants_dict, case_variant



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils, xes_constants, constants, pandas_utils
from pm4py.objects.conversion.log import converter as log_converter
from typing import Optional, Dict, Any, Union, List
from pm4py.objects.log.obj import EventLog
import pandas as pd


class Outputs(Enum):
    DFG = "dfg"
    SEQUENCE = "sequence"
    PARALLEL = "parallel"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    ACTIVITIES = "activities"
    SKIPPABLE = "skippable"
    ACTIVITIES_ALWAYS_HAPPENING = "activities_always_happening"
    MIN_TRACE_LENGTH = "min_trace_length"
    TRACE = "trace"


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ENABLE_ACT_ALWAYS_EXECUTED = "enable_act_always_executed"


class ConfOutputs(Enum):
    FOOTPRINTS = "footprints"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    ACTIVITIES_ALWAYS_HAPPENING = "activities_always_happening"
    MIN_LENGTH_FIT = "min_length_fit"
    IS_FOOTPRINTS_FIT = "is_footprints_fit"


def apply(
    log_footprints: List[Dict[str, Any]],
    model_footprints: Dict[str, Any],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[Dict[str, Any]]:
    """
    Apply footprints conformance between a log footprints object
    and a model footprints object

    Parameters
    -----------------
    log_footprints
        Footprints of the log (trace-by-trace)
    model_footprints
        Footprints of the model
    parameters
        Parameters of the algorithm

    Returns
    ------------------
    violations
        List containing, for each trace, a dictionary containing the violations
    """
    if parameters is None:
        parameters = {}

    if not type(log_footprints) is list:
        raise Exception(
            "it is possible to apply this variant only on trace-by-trace footprints, not overall log footprints!"
        )

    conf_traces = {}

    enable_act_always_executed = exec_utils.get_param_value(
        Parameters.ENABLE_ACT_ALWAYS_EXECUTED, parameters, True
    )
    model_configurations = model_footprints[Outputs.SEQUENCE.value].union(
        model_footprints[Outputs.PARALLEL.value]
    )

    ret = []
    for tr in log_footprints:
        trace = tr[Outputs.TRACE.value]
        if trace in conf_traces:
            ret.append(conf_traces[trace])
        else:
            trace_configurations = tr[Outputs.SEQUENCE.value].union(
                tr[Outputs.PARALLEL.value]
            )
            trace_violations = {}
            trace_violations[ConfOutputs.FOOTPRINTS.value] = set(
                x
                for x in trace_configurations
                if x not in model_configurations
            )
            trace_violations[ConfOutputs.START_ACTIVITIES.value] = (
                set(
                    x
                    for x in tr[Outputs.START_ACTIVITIES.value]
                    if x
                    not in model_footprints[Outputs.START_ACTIVITIES.value]
                )
                if Outputs.START_ACTIVITIES.value in model_footprints
                else set()
            )
            trace_violations[ConfOutputs.END_ACTIVITIES.value] = (
                set(
                    x
                    for x in tr[Outputs.END_ACTIVITIES.value]
                    if x not in model_footprints[Outputs.END_ACTIVITIES.value]
                )
                if Outputs.END_ACTIVITIES.value in model_footprints
                else set()
            )
            trace_violations[ConfOutputs.ACTIVITIES_ALWAYS_HAPPENING.value] = (
                set(
                    x
                    for x in model_footprints[
                        Outputs.ACTIVITIES_ALWAYS_HAPPENING.value
                    ]
                    if x not in tr[Outputs.ACTIVITIES.value]
                )
                if Outputs.ACTIVITIES_ALWAYS_HAPPENING.value
                in model_footprints
                and enable_act_always_executed
                else set()
            )
            trace_violations[ConfOutputs.MIN_LENGTH_FIT.value] = (
                tr[Outputs.MIN_TRACE_LENGTH.value]
                >= model_footprints[Outputs.MIN_TRACE_LENGTH.value]
                if Outputs.MIN_TRACE_LENGTH.value in tr
                and Outputs.MIN_TRACE_LENGTH.value in model_footprints
                else True
            )
            trace_violations[ConfOutputs.IS_FOOTPRINTS_FIT.value] = (
                len(trace_violations[ConfOutputs.FOOTPRINTS.value]) == 0
                and len(trace_violations[ConfOutputs.START_ACTIVITIES.value])
                == 0
                and len(trace_violations[ConfOutputs.END_ACTIVITIES.value])
                == 0
                and len(
                    trace_violations[
                        ConfOutputs.ACTIVITIES_ALWAYS_HAPPENING.value
                    ]
                )
                == 0
                and trace_violations[ConfOutputs.MIN_LENGTH_FIT.value]
            )

            ret.append(trace_violations)
            conf_traces[trace] = trace_violations

    return ret


def get_diagnostics_dataframe(
    log: EventLog,
    conf_result: List[Dict[str, Any]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pd.DataFrame:
    """
    Gets the diagnostics dataframe from the log
    and the results of footprints conformance checking
    (trace-by-trace)

    Parameters
    --------------
    log
        Event log
    conf_result
        Conformance checking results (trace-by-trace)

    Returns
    --------------
    diagn_dataframe
        Diagnostics dataframe
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, xes_constants.DEFAULT_TRACEID_KEY
    )

    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters)

    diagn_stream = []

    for index in range(len(log)):
        case_id = log[index].attributes[case_id_key]
        is_fit = conf_result[index][ConfOutputs.IS_FOOTPRINTS_FIT.value]
        footprints_violations = len(
            conf_result[index][ConfOutputs.FOOTPRINTS.value]
        )
        start_activities_violations = len(
            conf_result[index][ConfOutputs.START_ACTIVITIES.value]
        )
        end_activities_violations = len(
            conf_result[index][ConfOutputs.END_ACTIVITIES.value]
        )
        act_always_happening_violations = len(
            conf_result[index][ConfOutputs.ACTIVITIES_ALWAYS_HAPPENING.value]
        )
        min_length_fit = conf_result[index][ConfOutputs.MIN_LENGTH_FIT.value]

        diagn_stream.append(
            {
                "case_id": case_id,
                "is_fit": is_fit,
                "footprints_violations": footprints_violations,
                "start_activities_violations": start_activities_violations,
                "end_activities_violations": end_activities_violations,
                "act_always_happening_violations": act_always_happening_violations,
                "min_length_fit": min_length_fit,
            })

    return pandas_utils.instantiate_dataframe(diagn_stream)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.dfg import utils
from enum import Enum
from typing import Optional, Dict, Any, Tuple


class Outputs(Enum):
    DFG = "dfg"
    SEQUENCE = "sequence"
    PARALLEL = "parallel"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    ACTIVITIES = "activities"
    SKIPPABLE = "skippable"
    ACTIVITIES_ALWAYS_HAPPENING = "activities_always_happening"
    MIN_TRACE_LENGTH = "min_trace_length"
    TRACE = "trace"


def apply(
    dfg: Dict[Tuple[str, str], int],
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, Any]:
    """
    Discovers a footprint object from a DFG

    Parameters
    --------------
    dfg
        DFG
    parameters
        Parameters of the algorithm

    Returns
    --------------
    footprints_obj
        Footprints object
    """
    if parameters is None:
        parameters = {}

    parallel = {(x, y) for (x, y) in dfg if (y, x) in dfg}
    sequence = {(x, y) for (x, y) in dfg if not (y, x) in dfg}
    # replace this if needed
    start_activities = set(utils.dfg_utils.infer_start_activities(dfg))
    # replace this if needed
    end_activities = set(utils.dfg_utils.infer_end_activities(dfg))
    activities = set(utils.dfg_utils.get_activities_from_dfg(dfg))

    return {
        Outputs.SEQUENCE.value: sequence,
        Outputs.PARALLEL.value: parallel,
        Outputs.START_ACTIVITIES.value: start_activities,
        Outputs.END_ACTIVITIES.value: end_activities,
        Outputs.ACTIVITIES.value: activities,
    }



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union
import polars as pl
from pm4py.util import exec_utils, constants, xes_constants


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY


def apply(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, Dict[str, int]]:
    """
    Computes for each trace of the event log how much rework occurs.
    The rework is computed as the difference between the total number of activities of a trace and the
    number of unique activities.

    Parameters
    ----------------
    lf
        Polars LazyFrame
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => the activity key
        - Parameters.CASE_ID_KEY => the case identifier attribute

    Returns
    -----------------
    dict
        Dictionary associating to each case ID:
        - The number of total activities of the case (number of events)
        - The rework (difference between the total number of activities of a trace and the number of unique activities)
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    # Group by case and compute count and nunique for activities
    grouped_df = (
        lf.group_by(case_id_key)
        .agg([
            pl.col(activity_key).count().alias("count"),
            pl.col(activity_key).n_unique().alias("nunique")
        ])
        .collect()
    )

    rework_cases = {}
    for row in grouped_df.iter_rows():
        case_id = row[0]
        count = row[1]
        nunique = row[2]
        rework_cases[case_id] = {
            "number_activities": count,
            "rework": count - nunique,
        }

    return rework_cases


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.business_hours import BusinessHours
from pm4py.objects.log.util import sorting
from pm4py.util import constants
from pm4py.util import xes_constants as xes
from pm4py.objects.log.obj import EventLog, Trace, Event
from copy import copy
from enum import Enum
from pm4py.util import exec_utils


class Parameters(Enum):
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TRANSITION_KEY = constants.PARAMETER_CONSTANT_TRANSITION_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    LIFECYCLE_INSTANCE_KEY = "pm4py:param:lifecycle:instance:key"
    BUSINESS_HOURS = "business_hours"
    BUSINESS_HOUR_SLOTS = "business_hour_slots"
    WORKCALENDAR = "workcalendar"


def to_interval(log, parameters=None):
    """
    Converts a log to interval format (e.g. an event has two timestamps)
    from lifecycle format (an event has only a timestamp, and a transition lifecycle)

    Parameters
    -------------
    log
        Log (expressed in the lifecycle format)
    parameters
        Possible parameters of the method (activity, timestamp key, start timestamp key, transition ...)

    Returns
    -------------
    log
        Interval event log
    """
    if parameters is None:
        parameters = {}

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, xes.DEFAULT_TIMESTAMP_KEY
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes.DEFAULT_START_TIMESTAMP_KEY,
    )
    transition_key = exec_utils.get_param_value(
        Parameters.TRANSITION_KEY, parameters, xes.DEFAULT_TRANSITION_KEY
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )
    lifecycle_instance_key = exec_utils.get_param_value(
        Parameters.LIFECYCLE_INSTANCE_KEY, parameters, xes.DEFAULT_INSTANCE_KEY
    )
    business_hours = exec_utils.get_param_value(
        Parameters.BUSINESS_HOURS, parameters, False
    )
    business_hours_slots = exec_utils.get_param_value(
        Parameters.BUSINESS_HOUR_SLOTS,
        parameters,
        constants.DEFAULT_BUSINESS_HOUR_SLOTS,
    )

    if log is not None and len(log) > 0:
        if (
            "PM4PY_TYPE" in log.attributes
            and log.attributes["PM4PY_TYPE"] == "interval"
        ):
            return log
        if log[0] is not None and len(log[0]) > 0:
            first_event = log[0][0]
            if start_timestamp_key in first_event:
                return log

        new_log = EventLog(
            attributes=copy(log.attributes),
            extensions=copy(log.extensions),
            classifiers=copy(log.classifiers),
            omni_present=copy(log.omni_present),
            properties=copy(log.properties),
        )
        new_log.attributes["PM4PY_TYPE"] = "interval"
        new_log.properties[
            constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
        ] = xes.DEFAULT_START_TIMESTAMP_KEY

        for trace in log:
            new_trace = Trace()
            for attr in trace.attributes:
                new_trace.attributes[attr] = trace.attributes[attr]
            activities_start = {}
            for event in trace:
                activity = event[activity_key]
                instance = (
                    event[lifecycle_instance_key]
                    if lifecycle_instance_key in event
                    else None
                )
                activity = (activity, instance)
                transition = (
                    event[transition_key]
                    if transition_key in event
                    else "complete"
                )
                timestamp = event[timestamp_key]
                if transition.lower() == "start":
                    if activity not in activities_start:
                        activities_start[activity] = list()
                    activities_start[activity].append(event)
                elif transition.lower() == "complete":
                    start_event = None
                    start_timestamp = event[timestamp_key]
                    if (
                        activity in activities_start
                        and len(activities_start[activity]) > 0
                    ):
                        start_event = activities_start[activity].pop(0)
                        start_timestamp = start_event[timestamp_key]
                    new_event = Event()
                    for attr in event:
                        if (
                            not attr == timestamp_key
                            and not attr == transition_key
                        ):
                            new_event[attr] = event[attr]
                    if start_event is not None:
                        for attr in start_event:
                            if (
                                not attr == timestamp_key
                                and not attr == transition_key
                            ):
                                new_event["@@startevent_" + attr] = (
                                    start_event[attr]
                                )
                    new_event[start_timestamp_key] = start_timestamp
                    new_event[timestamp_key] = timestamp
                    new_event["@@duration"] = (
                        timestamp - start_timestamp
                    ).total_seconds()

                    if business_hours:
                        bh = BusinessHours(
                            start_timestamp,
                            timestamp,
                            business_hour_slots=business_hours_slots,
                        )
                        new_event["@@approx_bh_duration"] = bh.get_seconds()

                    new_trace.append(new_event)
            new_trace = sorting.sort_timestamp_trace(
                new_trace, start_timestamp_key
            )
            new_log.append(new_trace)
        return new_log

    return log


def to_lifecycle(log, parameters=None):
    """
    Converts a log from interval format (e.g. an event has two timestamps)
    to lifecycle format (an event has only a timestamp, and a transition lifecycle)

    Parameters
    -------------
    log
        Log (expressed in the interval format)
    parameters
        Possible parameters of the method (activity, timestamp key, start timestamp key, transition ...)

    Returns
    -------------
    log
        Lifecycle event log
    """
    if parameters is None:
        parameters = {}

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, xes.DEFAULT_TIMESTAMP_KEY
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes.DEFAULT_START_TIMESTAMP_KEY,
    )
    transition_key = exec_utils.get_param_value(
        Parameters.TRANSITION_KEY, parameters, xes.DEFAULT_TRANSITION_KEY
    )

    if log is not None and len(log) > 0:
        if (
            "PM4PY_TYPE" in log.attributes
            and log.attributes["PM4PY_TYPE"] == "lifecycle"
        ):
            return log
        if log[0] is not None and len(log[0]) > 0:
            first_event = log[0][0]
            if transition_key in first_event:
                return log

        new_log = EventLog(
            attributes=copy(log.attributes),
            extensions=copy(log.extensions),
            classifiers=copy(log.classifiers),
            omni_present=copy(log.omni_present),
            properties=copy(log.properties),
        )
        new_log.attributes["PM4PY_TYPE"] = "lifecycle"

        for trace in log:
            new_trace = Trace()
            for attr in trace.attributes:
                new_trace.attributes[attr] = trace.attributes[attr]
            list_events = []
            for index, event in enumerate(trace):
                new_event_start = Event()
                new_event_complete = Event()
                for attr in event:
                    if (
                        not attr == timestamp_key
                        and not attr == start_timestamp_key
                    ):
                        new_event_start[attr] = event[attr]
                        new_event_complete[attr] = event[attr]
                new_event_start[timestamp_key] = event[start_timestamp_key]
                new_event_start[transition_key] = "start"
                new_event_start["@@custom_lif_id"] = 0
                new_event_start["@@origin_ev_idx"] = index
                new_event_complete[timestamp_key] = event[timestamp_key]
                new_event_complete[transition_key] = "complete"
                new_event_complete["@@custom_lif_id"] = 1
                new_event_complete["@@origin_ev_idx"] = index
                list_events.append(new_event_start)
                list_events.append(new_event_complete)
            list_events = sorted(
                list_events,
                key=lambda x: (
                    x[timestamp_key],
                    x["@@origin_ev_idx"],
                    x["@@custom_lif_id"],
                ),
            )
            for ev in list_events:
                new_trace.append(ev)
            new_log.append(new_trace)
        return new_log
    return log


def assign_lead_cycle_time(log, parameters=None):
    """
    Assigns the lead and cycle time to an interval log

    Parameters
    -------------
    log
        Interval log
    parameters
        Parameters of the algorithm, including: start_timestamp_key, timestamp_key, business_hour_slots
    """
    if parameters is None:
        parameters = {}

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, xes.DEFAULT_TIMESTAMP_KEY
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes.DEFAULT_START_TIMESTAMP_KEY,
    )
    business_hours_slots = exec_utils.get_param_value(
        Parameters.BUSINESS_HOUR_SLOTS,
        parameters,
        constants.DEFAULT_BUSINESS_HOUR_SLOTS,
    )

    interval_log = to_interval(log, parameters=parameters)

    for trace in interval_log:
        approx_partial_lead_time = 0
        approx_partial_cycle_time = 0
        approx_wasted_time = 0
        max_et = None
        max_et_seconds = 0
        for i in range(len(trace)):
            this_wasted_time = 0
            st = trace[i][start_timestamp_key]
            st_seconds = st.timestamp()
            et = trace[i][timestamp_key]
            et_seconds = et.timestamp()

            if max_et_seconds > 0 and st_seconds > max_et_seconds:
                bh_unworked = BusinessHours(
                    max_et, st, business_hour_slots=business_hours_slots
                )
                unworked_sec = bh_unworked.get_seconds()
                approx_partial_lead_time = (
                    approx_partial_lead_time + unworked_sec
                )
                approx_wasted_time = approx_wasted_time + unworked_sec
                this_wasted_time = unworked_sec

            if st_seconds > max_et_seconds:
                bh = BusinessHours(
                    st, et, business_hour_slots=business_hours_slots
                )
                approx_bh_duration = bh.get_seconds()

                approx_partial_cycle_time = (
                    approx_partial_cycle_time + approx_bh_duration
                )
                approx_partial_lead_time = (
                    approx_partial_lead_time + approx_bh_duration
                )
            elif st_seconds < max_et_seconds and et_seconds > max_et_seconds:
                bh = BusinessHours(
                    max_et, et, business_hour_slots=business_hours_slots
                )
                approx_bh_duration = bh.get_seconds()

                approx_partial_cycle_time = (
                    approx_partial_cycle_time + approx_bh_duration
                )
                approx_partial_lead_time = (
                    approx_partial_lead_time + approx_bh_duration
                )

            if et_seconds > max_et_seconds:
                max_et_seconds = et_seconds
                max_et = et

            ratio_cycle_lead_time = 1
            if approx_partial_lead_time > 0:
                ratio_cycle_lead_time = (
                    approx_partial_cycle_time / approx_partial_lead_time
                )

            trace[i][
                "@@approx_bh_partial_cycle_time"
            ] = approx_partial_cycle_time
            trace[i][
                "@@approx_bh_partial_lead_time"
            ] = approx_partial_lead_time
            trace[i]["@@approx_bh_overall_wasted_time"] = approx_wasted_time
            trace[i]["@@approx_bh_this_wasted_time"] = this_wasted_time
            trace[i][
                "@approx_bh_ratio_cycle_lead_time"
            ] = ratio_cycle_lead_time

    return interval_log



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
def search_act_class_attr(log, force_activity_transition_insertion=False):
    """
    Search among classifiers expressed in the log one that is good for the process model extraction

    Parameters
    -----------
    log
        Trace log
    force_activity_transition_insertion
        Optionally force the activitiy+transition classifier insertion

    Returns
    -----------
    log
        Trace log (plus eventually one additional event attribute as the classifier)
    """
    classifier = None
    if (
        log.classifiers
        and "Activity classifier" in log.classifiers
        and log.classifiers["Activity classifier"]
    ):
        classifier = "Activity classifier"
    elif (
        log.classifiers
        and "MXML Legacy Classifier" in log.classifiers
        and log.classifiers["MXML Legacy Classifier"]
    ):
        classifier = "MXML Legacy Classifier"
    return insert_activity_classifier_attribute(
        log,
        classifier,
        force_activity_transition_insertion=force_activity_transition_insertion,
    )


def insert_activity_classifier_attribute(
    log, classifier, force_activity_transition_insertion=False
):
    """
    Insert the specified classifier as additional event attribute in the log

    Parameters
    -----------
    log
        Trace log
    classifier
        Event classifier
    force_activity_transition_insertion
        Optionally force the activitiy+transition classifier insertion

    Returns
    --------
    log
        Trace log (plus eventually one additional event attribute as the classifier)
    classifier_attr_key
        Attribute name of the attribute that contains the classifier value
    """
    classifier_attr_key = None
    if classifier is not None:
        classifier_attr_key = "@@classifier"
        for trace in log:
            for event in trace:
                classifier_value = "+".join(
                    [event[x] for x in log.classifiers[classifier]]
                )
                event[classifier_attr_key] = classifier_value
    elif force_activity_transition_insertion:
        if (
            len(log) > 0
            and len(log[0]) > 0
            and "concept:name" in log[0][0]
            and "lifecycle:transition" in log[0][0]
        ):
            classifier_attr_key = "@@classifier"
            for trace in log:
                for event in trace:
                    classifier_value = (
                        event["concept:name"]
                        + "+"
                        + event["lifecycle:transition"]
                    )
                    event[classifier_attr_key] = classifier_value
    return log, classifier_attr_key


def insert_trace_classifier_attribute(log, classifier):
    """
    Insert the specified classifier as additional trace attribute in the log

    Parameter
    -----------
    log
        Trace log
    classifier
        Event classifier

    Returns
    -----------
    log
        Trace log (plus eventually one additional event attribute as the classifier)
    classifier_attr_key
        Attribute name of the attribute that contains the classifier value
    """
    classifier_attr_key = None
    if classifier is not None:
        classifier_attr_key = "@@traceClassifier"
        for trace in log:
            classifier_value = "+".join(
                [trace.attributes[x] for x in log.classifiers[classifier]]
            )
            trace.attributes[classifier_attr_key] = classifier_value
    return log, classifier_attr_key



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.exporter.jsonocel.variants import ocel20
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from enum import Enum
import json
from pm4py.util import exec_utils, constants as pm4_constants


class Parameters(Enum):
    ENCODING = "encoding"


def apply(
    ocel: OCEL, target_path: str, parameters: Optional[Dict[Any, Any]] = None
):
    """
    Exports an object-centric event log (OCEL 2.0) in a JSON-OCEL 2.0 standard file

    Parameters
    ------------------
    ocel
        Object-centric event log
    target_path
        Destination path
    parameters
        Possible parameters of the method, including:
        - Parameters.ENCODING
    """
    if parameters is None:
        parameters = {}

    encoding = exec_utils.get_param_value(
        Parameters.ENCODING, parameters, pm4_constants.DEFAULT_ENCODING
    )

    legacy_object = ocel20.get_enriched_object(ocel)

    json_object = {}
    json_object["objectTypes"] = []
    json_object["eventTypes"] = []
    json_object["objects"] = []
    json_object["events"] = []

    for ot, attrs in legacy_object["ocel:objectTypes"].items():
        descr = {
            "name": ot,
            "attributes": [{"name": x, "type": y} for x, y in attrs.items()],
        }
        json_object["objectTypes"].append(descr)

    for et, attrs in legacy_object["ocel:eventTypes"].items():
        descr = {
            "name": et,
            "attributes": [{"name": x, "type": y} for x, y in attrs.items()],
        }
        json_object["eventTypes"].append(descr)

    obj_idx = {}
    for objid, obj in legacy_object["ocel:objects"].items():
        descr = {}
        descr["id"] = objid
        descr["type"] = obj["ocel:type"]
        if "ocel:ovmap" in obj and obj["ocel:ovmap"]:
            descr["attributes"] = []
            for k, v in obj["ocel:ovmap"].items():
                descr["attributes"].append(
                    {"name": k, "time": "1970-01-01T00:00:00Z", "value": v}
                )
        if "ocel:o2o" in obj and obj["ocel:o2o"]:
            descr["relationships"] = []
            for v in obj["ocel:o2o"]:
                descr["relationships"].append(
                    {
                        "objectId": v["ocel:oid"],
                        "qualifier": v["ocel:qualifier"],
                    }
                )
        json_object["objects"].append(descr)
        obj_idx[objid] = len(obj_idx)

    eve_idx = {}
    for evid, eve in legacy_object["ocel:events"].items():
        descr = {}
        descr["id"] = evid
        descr["type"] = eve["ocel:activity"]
        descr["time"] = eve["ocel:timestamp"]
        if "ocel:vmap" in eve and eve["ocel:vmap"]:
            descr["attributes"] = []
            for k, v in eve["ocel:vmap"].items():
                descr["attributes"].append({"name": k, "value": v})
        if "ocel:typedOmap" in eve and eve["ocel:typedOmap"]:
            descr["relationships"] = []
            for v in eve["ocel:typedOmap"]:
                descr["relationships"].append(
                    {
                        "objectId": v["ocel:oid"],
                        "qualifier": v["ocel:qualifier"],
                    }
                )
        json_object["events"].append(descr)
        eve_idx[evid] = len(eve_idx)

    for change in legacy_object["ocel:objectChanges"]:
        oid = change["ocel:oid"]
        obj = json_object["objects"][obj_idx[oid]]

        obj["attributes"].append(
            {
                "name": change["ocel:field"],
                "time": change["ocel:timestamp"],
                "value": change[change["ocel:field"]],
            }
        )

        json_object["objects"][obj_idx[oid]] = obj

    F = open(target_path, "w", encoding=encoding)
    json.dump(json_object, F, indent=2)
    F.close()



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from threading import Thread
from typing import Callable, Any, List, Optional
from pm4py.util import constants


class Pm4pyThreadManager:
    """
    A lightweight manager to optionally run submitted callables in threads.

    Args:
        is_threaded (bool): If True, `submit` runs the function in a new thread
                            and `join` waits for all threads to finish.
                            If False, `submit` runs synchronously and `join` does nothing.
    """
    def __init__(self, is_threaded: Optional[bool] = None) -> None:
        if is_threaded is None:
            is_threaded = constants.DEFAULT_IS_THREADING_MANAGEMENT_ENABLED
        self.is_threaded = is_threaded
        self._threads: List[Thread] = []

    def submit(self, func: Callable[..., Any], *args: Any, **kwargs: Any) -> Optional[Any]:
        """
        Submit a function to run with the provided args/kwargs.

        - If `is_threaded` is True, starts a thread and returns None immediately.
        - If `is_threaded` is False, runs synchronously and returns the function's result.
        """
        if self.is_threaded:
            t = Thread(target=func, args=args, kwargs=kwargs)
            self._threads.append(t)
            t.start()
            return None
        else:
            return func(*args, **kwargs)

    def join(self) -> None:
        """
        Wait for all threads to complete when `is_threaded` is True.
        Does nothing when `is_threaded` is False.
        """
        if not self.is_threaded:
            return
        for t in self._threads:
            t.join()
        self._threads.clear()



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Dict, Union, Any, Optional, List
from pm4py.objects.log.obj import EventLog
import pandas as pd
from enum import Enum
from pm4py.util import exec_utils
from pm4py.statistics.chaotic_activities.variants import niek_sidorova


class Variants(Enum):
    NIEK_SIDOROVA = niek_sidorova


def apply(log: Union[pd.DataFrame, EventLog], variant=Variants.NIEK_SIDOROVA,
          parameters: Optional[Dict[Any, Any]] = None) -> List[Dict[str, Any]]:
    """
    Compute metrics used to detect *chaotic activities* in an event log.

    Parameters
    -----------------
    log
        Event log or Pandas dataframe
    variant
        Variant of the algorithm: Variants.NIEK_SIDOROVA
    parameters
        Variant-specific parameters

    Returns
    -----------------
    chaotic_activities
        List of dictionaries, each representing an activity, sorted decreasingly based on the chaotic score (less is better).
    """
    return exec_utils.get_variant(variant).apply(log, parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import random


class ObjMarking(dict):
    def __init__(self, *args, **kwargs):
        self.picked_items = set()
        self.update(*args, **kwargs)

    def __getitem_original__(self, key):
        if key in self:
            return dict.__getitem__(self, key)
        return {}

    def __getitem__(self, key):
        return ObjMarkingAccessor(self, key)

    def __setitem__(self, key, val):
        dict.__setitem__(self, key, val)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v


class ObjMarkingAccessor(object):
    def __init__(self, obj_marking, key):
        self.obj_marking = obj_marking
        self.key = key

    def __sub__(self, other):
        if isinstance(other, int):
            picked_items = set(
                random.sample(
                    self.obj_marking.__getitem_original__(self.key),
                    min(
                        len(self.obj_marking.__getitem_original__(self.key)),
                        other,
                    ),
                )
            )
            self.obj_marking.picked_items = picked_items
            diff = self.obj_marking.__getitem_original__(self.key).difference(
                picked_items
            )
            return diff
        else:
            diff = self.obj_marking.__getitem_original__(self.key).difference(
                other
            )
            return diff

    def __add__(self, other):
        if isinstance(other, int):
            picked_items = set(
                random.sample(self.obj_marking.picked_items, other)
            )
            if self.key not in self.obj_marking:
                self.obj_marking[self.key] = set()
            new_value = self.obj_marking.__getitem_original__(self.key).union(
                picked_items
            )
            self.obj_marking[self.key] = new_value
            return new_value
        else:
            new_value = self.obj_marking.__getitem_original__(self.key).union(
                other
            )
            self.obj_marking[self.key] = new_value
            return new_value

    def __repr__(self):
        return str(self.obj_marking.__getitem_original__(self.key))

    def __str__(self):
        return str(self.obj_marking.__getitem_original__(self.key))

    def __lt__(self, other):
        if isinstance(other, int):
            if len(self.obj_marking.__getitem_original__(self.key)) < other:
                return True
            return False

    def __le__(self, other):
        if isinstance(other, int):
            if len(self.obj_marking.__getitem_original__(self.key)) <= other:
                return True
            return False

    def __gt__(self, other):
        if isinstance(other, int):
            if len(self.obj_marking.__getitem_original__(self.key)) > other:
                return True
            return False

    def __ge__(self, other):
        if isinstance(other, int):
            if len(self.obj_marking.__getitem_original__(self.key)) >= other:
                return True
            return False

    def __lt__(self, other):
        if isinstance(other, int):
            if len(self.obj_marking.__getitem_original__(self.key)) < other:
                return True
            return False

    def __eq__(self, other):
        if isinstance(other, int):
            if len(self.obj_marking.__getitem_original__(self.key)) <= other:
                return True
            return False
        else:
            return self.__dict__ == other.__dict__



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union
import polars as pl
from pm4py.util import constants, xes_constants, exec_utils


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY


def apply(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, int]:
    """
    Associates to each activity (with at least one rework) the number of cases in the log for which
    the rework happened.

    Parameters
    ------------------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => the attribute to be used as activity
        - Parameters.CASE_ID_KEY => the attribute to be used as case ID

    Returns
    ------------------
    dict
        Dictionary associating to each activity the number of cases for which the rework happened
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    # Select only needed columns
    selected_lf = lf.select([activity_key, case_id_key])
    
    # Add cumulative count for each activity within each case
    with_counts = (
        selected_lf
        .with_columns(
            pl.col(activity_key)
            .cum_count()
            .over([case_id_key, activity_key])
            .alias("__rework_count")
        )
    )
    
    # Filter for rework cases (count > 0, meaning activity occurred more than once)
    rework_cases = with_counts.filter(pl.col("__rework_count") > 1)
    
    # Get unique activity-case pairs for rework
    unique_rework = rework_cases.select([activity_key, case_id_key]).unique()
    
    # Count cases with rework per activity
    result_df = unique_rework.group_by(activity_key).count().collect()
    
    # Convert to dictionary
    ret = dict(zip(result_df[activity_key].to_list(), result_df["count"].to_list()))
    
    return ret


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.simulation.playout.petri_net.variants import extensive
from pm4py.algo.simulation.playout.petri_net.variants import (
    stochastic_playout,
    basic_playout,
)
from pm4py.util import exec_utils
from enum import Enum
from pm4py.objects.petri_net.obj import PetriNet, Marking
from typing import Optional, Dict, Any
from pm4py.objects.log.obj import EventLog


class Variants(Enum):
    BASIC_PLAYOUT = basic_playout
    STOCHASTIC_PLAYOUT = stochastic_playout
    EXTENSIVE = extensive


DEFAULT_VARIANT = Variants.BASIC_PLAYOUT
VERSIONS = {
    Variants.BASIC_PLAYOUT,
    Variants.EXTENSIVE,
    Variants.STOCHASTIC_PLAYOUT,
}


def apply(
    net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking = None,
    parameters: Optional[Dict[Any, Any]] = None,
    variant=DEFAULT_VARIANT,
) -> EventLog:
    """
    Do the playout of a Petrinet generating a log

    Parameters
    -----------
    net
        Petri net to play-out
    initial_marking
        Initial marking of the Petri net
    final_marking
        (if provided) Final marking of the Petri net
    parameters
        Parameters of the algorithm
    variant
        Variant of the algorithm to use:
            - Variants.BASIC_PLAYOUT: selects random traces from the model, without looking at the
            frequency of the transitions
            - Variants.STOCHASTIC_PLAYOUT: selects random traces from the model, looking at the
            stochastic frequency of the transitions. Requires the provision of the stochastic map
            or the log.
            - Variants.EXTENSIVE: gets all the traces from the model. can be expensive
    """
    return exec_utils.get_variant(variant).apply(
        net,
        initial_marking,
        final_marking=final_marking,
        parameters=parameters,
    )
